import { URLExt } from "@jupyterlab/coreutils";
import { ServerConnection } from "@jupyterlab/services";
import { PromiseDelegate } from "@lumino/coreutils";
import { Signal } from "@lumino/signaling";
import { Conda } from "./tokens";
/**
 * Conda Environment Manager
 */
export class CondaEnvironments {
    constructor(settings) {
        // Resolve promise to disconnect signals at disposal
        this._clean = () => {
            return;
        };
        this._isDisposed = false;
        this._environmentChanged = new Signal(this);
        this._environmentsTimer = -1;
        this._environmentTypes = {};
        this._fromHistory = false;
        this._packageManager = new CondaPackage();
        this._whitelist = false;
        this._environments = new Array();
        if (settings) {
            this._updateSettings(settings);
            settings.changed.connect(this._updateSettings, this);
            const clean = new Promise(((resolve) => {
                this._clean = resolve;
            }).bind(this));
            clean.then(() => {
                settings.changed.disconnect(this._updateSettings, this);
            });
        }
    }
    get environments() {
        return this.refresh().then(() => {
            return Promise.resolve(this._environments);
        });
    }
    get environmentChanged() {
        return this._environmentChanged;
    }
    getPackageManager(name) {
        this._packageManager.environment = name;
        return this._packageManager;
    }
    /**
     * Test whether the manager is disposed.
     */
    get isDisposed() {
        return this._isDisposed;
    }
    /**
     * Get the list of packages to install for a given environment type.
     *
     * The returned package list is the input type split with " ".
     *
     * @param type Environment type
     * @returns List of packages to create the environment
     */
    getEnvironmentFromType(type) {
        if (type in this._environmentTypes) {
            return this._environmentTypes[type];
        }
        return type.split(" ");
    }
    /**
     * Get the list of user-defined environment types
     */
    get environmentTypes() {
        return Object.keys(this._environmentTypes);
    }
    /**
     * Load user settings
     *
     * @param settings User settings
     */
    _updateSettings(settings) {
        this._environmentTypes = settings.get("types").composite;
        this._fromHistory = settings.get("fromHistory").composite;
        this._whitelist = settings.get("whitelist").composite;
    }
    /**
     * Dispose of the resources used by the manager.
     */
    dispose() {
        if (this.isDisposed) {
            return;
        }
        this._isDisposed = true;
        clearInterval(this._environmentsTimer);
        this._clean();
        this._environments.length = 0;
    }
    async getChannels(name) {
        try {
            const request = {
                method: "GET"
            };
            const { promise } = Private.requestServer(URLExt.join("conda", "channels"), request);
            const response = await promise;
            if (response.ok) {
                const data = await response.json();
                return data["channels"];
            }
        }
        catch (error) {
            let message = error.message || error.toString();
            if (message !== "cancelled") {
                console.error(message);
                message = `Fail to get the channels for environment ${name}.`;
            }
            throw new Error(message);
        }
    }
    async clone(target, name) {
        try {
            const request = {
                body: JSON.stringify({ name, twin: target }),
                method: "POST"
            };
            const { promise } = Private.requestServer(URLExt.join("conda", "environments"), request);
            const response = await promise;
            if (response.ok) {
                this._environmentChanged.emit({
                    name: name,
                    source: target,
                    type: "clone"
                });
            }
        }
        catch (error) {
            let message = error.message || error.toString();
            if (message !== "cancelled") {
                console.error(message);
                message = `An error occurred while cloning environment "${target}".`;
            }
            throw new Error(message);
        }
    }
    async create(name, type) {
        try {
            const packages = this.getEnvironmentFromType(type || "");
            const request = {
                body: JSON.stringify({ name, packages }),
                method: "POST"
            };
            const { promise } = Private.requestServer(URLExt.join("conda", "environments"), request);
            const response = await promise;
            if (response.ok) {
                this._environmentChanged.emit({
                    name: name,
                    source: packages,
                    type: "create"
                });
            }
        }
        catch (error) {
            let message = error.message || error.toString();
            if (message !== "cancelled") {
                console.error(message);
                message = `An error occurred while creating environment "${name}".`;
            }
            throw new Error(message);
        }
    }
    export(name, fromHistory = null) {
        if (fromHistory === null) {
            fromHistory = this._fromHistory;
        }
        try {
            const request = {
                method: "GET"
            };
            const args = URLExt.objectToQueryString({
                download: 1,
                history: fromHistory ? 1 : 0
            });
            const { promise } = Private.requestServer(URLExt.join("conda", "environments", name) + args, request);
            return promise;
        }
        catch (error) {
            let message = error.message || error.toString();
            if (message !== "cancelled") {
                console.error(message);
                message = `An error occurred while exporting environment "${name}".`;
            }
            throw new Error(message);
        }
    }
    async import(name, fileContent, fileName) {
        try {
            const data = { name, file: fileContent };
            if (fileName) {
                data["filename"] = fileName;
            }
            const request = {
                body: JSON.stringify(data),
                method: "POST"
            };
            const { promise } = Private.requestServer(URLExt.join("conda", "environments"), request);
            const response = await promise;
            if (response.ok) {
                this._environmentChanged.emit({
                    name: name,
                    source: fileContent,
                    type: "import"
                });
            }
        }
        catch (error) {
            let message = error.message || error.toString();
            if (message !== "cancelled") {
                console.error(message);
                message = `An error occurred while importing "${name}".`;
            }
            throw new Error(message);
        }
    }
    async refresh() {
        try {
            const request = {
                method: "GET"
            };
            const queryArgs = {
                whitelist: this._whitelist ? 1 : 0
            };
            const { promise } = Private.requestServer(URLExt.join("conda", "environments") +
                URLExt.objectToQueryString(queryArgs), request);
            const response = await promise;
            const data = (await response.json());
            this._environments = data.environments;
            return data.environments;
        }
        catch (error) {
            let message = error.message || error.toString();
            if (message !== "cancelled") {
                console.error(message);
                message = "An error occurred while listing Conda environments.";
            }
            throw new Error(message);
        }
    }
    async remove(name) {
        try {
            const request = {
                method: "DELETE"
            };
            const { promise } = await Private.requestServer(URLExt.join("conda", "environments", name), request);
            const response = await promise;
            if (response.ok) {
                this._environmentChanged.emit({
                    name: name,
                    source: null,
                    type: "remove"
                });
            }
        }
        catch (error) {
            let message = error.message || error.toString();
            if (message !== "cancelled") {
                console.error(message);
                message = `An error occurred while removing "${name}".`;
            }
            throw new Error(message);
        }
    }
    async update(name, fileContent, fileName) {
        try {
            const data = { file: fileContent };
            if (fileName) {
                data["filename"] = fileName;
            }
            const request = {
                body: JSON.stringify(data),
                method: "PATCH"
            };
            const { promise } = Private.requestServer(URLExt.join("conda", "environments", name), request);
            const response = await promise;
            if (response.ok) {
                this._environmentChanged.emit({
                    name: name,
                    source: fileContent,
                    type: "update"
                });
            }
        }
        catch (error) {
            let message = error.message || error.toString();
            if (message !== "cancelled") {
                console.error(message);
                message = `An error occurred while updating "${name}".`;
            }
            throw new Error(message);
        }
    }
}
export class CondaPackage {
    constructor(environment) {
        this._packageChanged = new Signal(this);
        this._cancellableStack = [];
        this.environment = environment;
    }
    get packageChanged() {
        return this._packageChanged;
    }
    /**
     * Refresh the package list.
     *
     * @param includeAvailable Include available package list
     * @param environment Environment name
     *
     * @returns The package list
     */
    async refresh(includeAvailable = true, environment) {
        const theEnvironment = environment || this.environment;
        if (theEnvironment === undefined) {
            return Promise.resolve([]);
        }
        this._cancelTasks();
        try {
            const request = {
                method: "GET"
            };
            // Get installed packages
            const { promise, cancel } = Private.requestServer(URLExt.join("conda", "environments", theEnvironment), request);
            const idx = this._cancellableStack.push(cancel) - 1;
            const response = await promise;
            this._cancellableStack.splice(idx, 1);
            const data = (await response.json());
            const installedPkgs = data.packages;
            const allPackages = [];
            if (includeAvailable) {
                // Get all available packages
                allPackages.push(...(await this._getAvailablePackages()));
            }
            // Set installed package status
            //- packages are sorted by name, we take advantage of this.
            const finalList = [];
            let availableIdx = 0;
            let installedIdx = 0;
            while (installedIdx < installedPkgs.length ||
                availableIdx < allPackages.length) {
                const installed = installedPkgs[installedIdx];
                let pkg = allPackages[availableIdx] || Object.assign(Object.assign({}, installed), { version: [installed.version], build_number: [installed.build_number], build_string: [installed.build_string], summary: "", home: "", keywords: "", tags: "" });
                pkg.summary = pkg.summary || "";
                // Stringify keywords and tags
                pkg.keywords = (pkg.keywords || "").toString().toLowerCase();
                pkg.tags = (pkg.tags || "").toString().toLowerCase();
                pkg.version_installed = "";
                pkg.version_selected = "none";
                pkg.updatable = false;
                if (installed !== undefined) {
                    if (pkg.name > installed.name) {
                        // installed is not in available
                        pkg = Object.assign(Object.assign({}, installed), { version: [installed.version], build_number: [installed.build_number], build_string: [installed.build_string], summary: "", home: "", keywords: "", tags: "" });
                        availableIdx -= 1;
                    }
                    if (pkg.name === installed.name) {
                        pkg.version_installed = installed.version;
                        pkg.version_selected = installed.version;
                        if (pkg.version.indexOf(installed.version) < 0) {
                            pkg.version.push(installed.version);
                        }
                        installedIdx += 1;
                    }
                }
                // Simplify the package channel name
                const splitUrl = pkg.channel.split("/");
                if (splitUrl.length > 2) {
                    let firstNotEmpty = 0;
                    if (["http:", "https:", "file:"].indexOf(splitUrl[firstNotEmpty]) >= 0) {
                        firstNotEmpty = 1; // Skip the scheme http, https or file
                    }
                    while (splitUrl[firstNotEmpty].length === 0) {
                        firstNotEmpty += 1;
                    }
                    pkg.channel = splitUrl[firstNotEmpty];
                    let pos = splitUrl.length - 1;
                    while (Conda.PkgSubDirs.indexOf(splitUrl[pos]) > -1 &&
                        pos > firstNotEmpty) {
                        pos -= 1;
                    }
                    if (pos > firstNotEmpty) {
                        pkg.channel += "/...";
                    }
                    pkg.channel += "/" + splitUrl[pos];
                }
                finalList.push(pkg);
                availableIdx += 1;
            }
            return finalList;
        }
        catch (error) {
            let message = error.message || error.toString();
            if (message !== "cancelled") {
                console.error(message);
                message = "An error occurred while retrieving available packages.";
            }
            throw new Error(message);
        }
    }
    async install(packages, environment) {
        const theEnvironment = environment || this.environment;
        if (theEnvironment === undefined || packages.length === 0) {
            return Promise.resolve();
        }
        try {
            const request = {
                body: JSON.stringify({ packages }),
                method: "POST"
            };
            const { promise } = Private.requestServer(URLExt.join("conda", "environments", theEnvironment, "packages"), request);
            const response = await promise;
            if (response.ok) {
                this._packageChanged.emit({
                    environment: theEnvironment,
                    type: "install",
                    packages
                });
            }
        }
        catch (error) {
            let message = error.message || error.toString();
            if (message !== "cancelled") {
                console.error(message);
                message = "An error occurred while installing packages.";
            }
            throw new Error(message);
        }
    }
    async develop(path, environment) {
        const theEnvironment = environment || this.environment;
        if (theEnvironment === undefined || path.length === 0) {
            return Promise.resolve();
        }
        try {
            const request = {
                body: JSON.stringify({ packages: [path] }),
                method: "POST"
            };
            const { promise } = Private.requestServer(URLExt.join("conda", "environments", theEnvironment, "packages") +
                URLExt.objectToQueryString({ develop: 1 }), request);
            const response = await promise;
            if (response.ok) {
                this._packageChanged.emit({
                    environment: theEnvironment,
                    type: "develop",
                    packages: [path]
                });
            }
        }
        catch (error) {
            let message = error.message || error.toString();
            if (message !== "cancelled") {
                console.error(message);
                message = `An error occurred while installing in development mode package in ${path}.`;
            }
            throw new Error(message);
        }
    }
    async check_updates(environment) {
        const theEnvironment = environment || this.environment;
        if (theEnvironment === undefined) {
            return Promise.resolve([]);
        }
        this._cancelTasks();
        try {
            const request = {
                method: "GET"
            };
            const { promise, cancel } = Private.requestServer(URLExt.join("conda", "environments", theEnvironment) +
                URLExt.objectToQueryString({ status: "has_update" }), request);
            const idx = this._cancellableStack.push(cancel) - 1;
            const response = await promise;
            this._cancellableStack.splice(idx, 1);
            const data = (await response.json());
            return data.updates.map(pkg => pkg.name);
        }
        catch (error) {
            let message = error.message || error.toString();
            if (message !== "cancelled") {
                console.error(message);
                message = "An error occurred while checking for package updates.";
            }
            throw new Error(message);
        }
    }
    async update(packages, environment) {
        const theEnvironment = environment || this.environment;
        if (theEnvironment === undefined) {
            return Promise.resolve();
        }
        try {
            const request = {
                body: JSON.stringify({ packages }),
                method: "PATCH"
            };
            const { promise } = Private.requestServer(URLExt.join("conda", "environments", theEnvironment, "packages"), request);
            const response = await promise;
            if (response.ok) {
                this._packageChanged.emit({
                    environment: theEnvironment,
                    type: "update",
                    packages
                });
            }
        }
        catch (error) {
            let message = error.message || error.toString();
            if (message !== "cancelled") {
                console.error(message);
                message = "An error occurred while updating packages.";
            }
            throw new Error(message);
        }
    }
    async remove(packages, environment) {
        const theEnvironment = environment || this.environment;
        if (theEnvironment === undefined) {
            return Promise.resolve();
        }
        try {
            const request = {
                body: JSON.stringify({ packages }),
                method: "DELETE"
            };
            const { promise } = Private.requestServer(URLExt.join("conda", "environments", theEnvironment, "packages"), request);
            const response = await promise;
            if (response.ok) {
                this._packageChanged.emit({
                    environment: theEnvironment,
                    type: "remove",
                    packages
                });
            }
        }
        catch (error) {
            let message = error.message || error.toString();
            if (message !== "cancelled") {
                console.error(message);
                message = "An error occurred while removing packages.";
            }
            throw new Error(message);
        }
    }
    async refreshAvailablePackages(cancellable = true) {
        await this._getAvailablePackages(true, cancellable);
    }
    /**
     * Does the available packages have description?
     *
     * @returns description presence
     */
    hasDescription() {
        return CondaPackage._hasDescription;
    }
    /**
     * Get the available packages list.
     *
     * @param force Force refreshing the available package list
     * @param cancellable Can this asynchronous action be cancelled?
     */
    async _getAvailablePackages(force = false, cancellable = true) {
        this._cancelTasks();
        if (CondaPackage._availablePackages === null || force) {
            const request = {
                method: "GET"
            };
            const { promise, cancel } = Private.requestServer(URLExt.join("conda", "packages"), request);
            let idx;
            if (cancellable) {
                idx = this._cancellableStack.push(cancel) - 1;
            }
            const response = await promise;
            if (idx) {
                this._cancellableStack.splice(idx, 1);
            }
            const data = (await response.json());
            CondaPackage._availablePackages = data.packages;
            CondaPackage._hasDescription = data.with_description || false;
        }
        return Promise.resolve(CondaPackage._availablePackages);
    }
    /**
     * Cancel optional tasks
     */
    _cancelTasks() {
        if (this._cancellableStack.length > 0) {
            const tasks = this._cancellableStack.splice(0, this._cancellableStack.length);
            tasks.forEach(task => {
                task();
            });
        }
    }
}
CondaPackage._availablePackages = null;
CondaPackage._hasDescription = false;
var Private;
(function (Private) {
    /**
     * Polling interval for accepted tasks
     */
    const POLLING_INTERVAL = 1000;
    /** Helper functions to carry on python notebook server request
     *
     * @param {string} url : request url
     * @param {RequestInit} request : initialization parameters for the request
     * @returns {ICancellablePromise<Response>} : Cancellable response to the request
     */
    Private.requestServer = function (url, request) {
        const settings = ServerConnection.makeSettings();
        const fullUrl = URLExt.join(settings.baseUrl, url);
        let answer;
        let cancelled = false;
        const promise = new PromiseDelegate();
        ServerConnection.makeRequest(fullUrl, request, settings)
            .then(response => {
            if (!response.ok) {
                response
                    .json()
                    .then(body => promise.reject(new ServerConnection.ResponseError(response, body.error)))
                    .catch(reason => {
                    console.error("Fail to read JSON response for request", request, reason);
                });
            }
            else if (response.status === 202) {
                const redirectUrl = response.headers.get("Location") || url;
                setTimeout((url, settings) => {
                    if (cancelled) {
                        // If cancelled, tell the backend to delete the task.
                        console.debug(`Request cancelled ${url}.`);
                        settings = Object.assign(Object.assign({}, settings), { method: "DELETE" });
                    }
                    answer = Private.requestServer(url, settings);
                    answer.promise
                        .then(response => promise.resolve(response))
                        .catch(reason => promise.reject(reason));
                }, POLLING_INTERVAL, redirectUrl, { method: "GET" });
            }
            else {
                promise.resolve(response);
            }
        })
            .catch(reason => {
            promise.reject(new ServerConnection.NetworkError(reason));
        });
        return {
            promise: promise.promise,
            cancel: () => {
                cancelled = true;
                if (answer) {
                    answer.cancel();
                }
                promise.reject("cancelled");
            }
        };
    };
})(Private || (Private = {}));
//# sourceMappingURL=services.js.map